/* global L, HTMLElement, HTMLFormElement, HTMLInputElement */
/* eslint import/no-unresolved: [2, {ignore: ['^https:\\/\\/unpkg\\.com\\/', '^https:\\/\\/cdn\\.jsdelivr\\.net\\/gh\\/locr-company\\/']}] */

import 'https://unpkg.com/leaflet@1.9.4/dist/leaflet.js';
import 'https://cdn.jsdelivr.net/gh/locr-company/Leaflet.GridLayer.VMS2@5241120/src/Leaflet.GridLayer.VMS2.js';
import Icon from '../../public/js/Icon.js';
import Settings from '../../public/js/Settings.js';

let eventsAlreadyLoaded = false;
let map;
let marker;

document.addEventListener( 'DOMContentLoaded', function () {
	const originalTbShow = window.tb_show;

	window.tb_show = function ( caption, url, imageGroup ) {
		originalTbShow( caption, url, imageGroup );

		let loadMapPromise;

		const urlObject = new URL( url );
		const urlHashSearchParams = new URLSearchParams( urlObject.hash );

		if ( urlHashSearchParams.has( 'inlineId' ) ) {
			if (
				urlHashSearchParams.get( 'inlineId' ) ===
				'onlinemap-location-overlay-content'
			) {
				const locationPreviewMap = document.getElementById(
					'location-preview-map'
				);
				if ( locationPreviewMap ) {
					loadMapPromise = loadMap( locationPreviewMap );
				}

				if ( ! eventsAlreadyLoaded ) {
					const locationAddressForm = document.getElementById(
						'location-address-form'
					);
					if ( locationAddressForm instanceof HTMLFormElement ) {
						locationAddressForm.addEventListener(
							'submit',
							( event ) => event.preventDefault()
						);
					}

					const locationAddressSearchButton = document.getElementById(
						'location-address-search'
					);
					if ( locationAddressSearchButton ) {
						locationAddressSearchButton.addEventListener(
							'click',
							() => {
								geocodeAddress()
									.then( setMapViewFromLocationResult )
									.then( setSelectedLocationIconToTheMap )
									.then( enableSaveLocationIconButton )
									.catch( handleExceptionAsLocationNotice );
							}
						);
					}
				}
			}
		}

		disableSaveLocationIconButton();
		if ( marker ) {
			marker.remove();
		}

		const locationIdElement = document.getElementById( 'location-id' );
		if ( locationIdElement instanceof HTMLInputElement ) {
			locationIdElement.value = '';
		}

		if ( ! eventsAlreadyLoaded ) {
			const saveLocationIconButton =
				document.getElementById( 'save-location-icon' );
			if ( saveLocationIconButton ) {
				saveLocationIconButton.addEventListener( 'click', async () => {
					getSelectedLocationIcon()
						.then( confirmCurrentLocation )
						.then( saveLocationIconToCurrentOnlineMapLocations )
						.then( closeThickboxAndReloadPage )
						.catch( handleExceptionAsLocationNotice );
				} );
			}

			const locationIconsElement =
				document.getElementById( 'location-icons' );
			if ( locationIconsElement instanceof HTMLElement ) {
				const locationIconCells =
					locationIconsElement.querySelectorAll( 'td' );
				for ( const locationIconCell of locationIconCells ) {
					locationIconCell.addEventListener( 'click', function () {
						for ( const cell of locationIconCells ) {
							cell.classList.remove( 'selected' );
						}
						this.classList.add( 'selected' );

						getSelectedLocationIcon().then(
							tryUpdateLocationIconOnTheMap
						);
					} );
				}
			}
		}

		if ( urlHashSearchParams.has( 'location-data' ) ) {
			try {
				const locationData = JSON.parse(
					urlHashSearchParams.get( 'location-data' )
				);
				if ( locationData.id ) {
					if ( locationIdElement instanceof HTMLInputElement ) {
						locationIdElement.value = locationData.id;
					}
				}
				if ( locationData.address ) {
					setAddressData( locationData.address );
				}
				if ( locationData.icon_id ) {
					setSelectedLocationIconByIconId( locationData.icon_id );
				}
				if ( locationData.position ) {
					const locationResult = {
						features: [
							{
								geometry: {
									coordinates: [
										locationData.position.longitude,
										locationData.position.latitude,
									],
									type: 'Point',
								},
							},
						],
					};
					loadMapPromise
						.then( () => {
							return setMapViewFromLocationResult(
								locationResult
							);
						} )
						.then( setSelectedLocationIconToTheMap )
						.then( enableSaveLocationIconButton )
						.catch( handleExceptionAsLocationNotice );
				}
			} catch ( exc ) {
				// Ignore error
			}
		}

		eventsAlreadyLoaded = true;
	};
} );

/**
 * @param {*} data
 * @return {Promise<*>} data
 */
function confirmCurrentLocation( data ) {
	return new Promise( ( resolve ) => {
		const dialog = document.createElement( 'dialog' );
		const p = document.createElement( 'p' );
		p.innerText = "Have you've been checked the current location?";
		dialog.appendChild( p );

		const form = document.createElement( 'form' );
		form.method = 'dialog';
		const yesButton = document.createElement( 'button' );
		yesButton.type = 'submit';
		yesButton.innerText = 'Yes';
		yesButton.value = 'yes';
		form.appendChild( yesButton );
		form.innerHTML += '&nbsp;&nbsp;';
		const cancelButton = document.createElement( 'button' );
		cancelButton.innerText = 'Cancel';
		cancelButton.value = 'cancel';
		form.appendChild( cancelButton );
		dialog.appendChild( form );

		document.body.appendChild( dialog );

		dialog.showModal();

		dialog.addEventListener( 'close', ( event ) => {
			const target = event.target;
			target.parentElement.removeChild( target );

			if ( target.returnValue === 'yes' ) {
				return resolve( data );
			}

			resolve( null );
		} );
	} );
}

/**
 * @param {Icon?} locationsIcon
 */
async function saveLocationIconToCurrentOnlineMapLocations( locationsIcon ) {
	if ( ! locationsIcon ) {
		return;
	}
	if ( ! marker ) {
		throw new Error( 'No marker found to add location icon!' );
	}

	const markerPosition = marker.getLatLng();

	const apiFetchHandler = window.wp?.apiFetch || window.wp?.apiRequest;
	if ( ! apiFetchHandler ) {
		throw new Error( 'No API fetch handler found!' );
	}

	const onlinemapIdElement = document.getElementById( 'onlinemap_id' );
	if ( ! ( onlinemapIdElement instanceof HTMLInputElement ) ) {
		throw new Error( 'No ID element found for the ONLINEmap!' );
	}

	const addressData = getAddressData();

	const urlParams = new URLSearchParams();
	urlParams.set( 'icon_id', locationsIcon.id );
	urlParams.set( 'latitude', markerPosition.lat );
	urlParams.set( 'longitude', markerPosition.lng );
	if ( addressData.street ) {
		urlParams.set( 'street', addressData.street );
	}
	if ( addressData.house ) {
		urlParams.set( 'house', addressData.house );
	}
	if ( addressData.city ) {
		urlParams.set( 'city', addressData.city );
	}
	if ( addressData.postal ) {
		urlParams.set( 'postal', addressData.postal );
	}
	if ( addressData.state ) {
		urlParams.set( 'state', addressData.state );
	}
	if ( addressData.country ) {
		urlParams.set( 'country', addressData.country );
	}

	const id = onlinemapIdElement.value;

	const method = 'POST';
	let path = `/locrmaps/onlinemap/${ id }/location`;

	const locationIdElement = document.getElementById( 'location-id' );
	if ( locationIdElement instanceof HTMLInputElement ) {
		const locationId = parseInt( locationIdElement.value );
		if ( ! isNaN( locationId ) ) {
			path = `/locrmaps/onlinemap/${ id }/location/${ locationId }`;
		}
	}

	return await apiFetchHandler( {
		path,
		method,
		headers: {
			'Content-Type': 'application/x-www-form-urlencoded',
		},
		body: urlParams,
	} );
}

/**
 * @param {*}    locationResult
 * @param {Icon} locationIcon
 */
async function buildMarkerFromLocationResultAndLocationIcon(
	locationResult,
	locationIcon
) {
	if (
		locationResult.features instanceof Array &&
		locationResult.features.length > 0
	) {
		const location = locationResult.features[ 0 ];
		return window.L.marker(
			[
				location.geometry.coordinates[ 1 ],
				location.geometry.coordinates[ 0 ],
			],
			{
				draggable: true,
				icon: locationIcon.toLeafletIcon(),
			}
		);
	}

	throw new Error( 'No location found to build marker!' );
}

async function closeThickboxAndReloadPage( data ) {
	if ( ! data ) {
		return;
	}

	window.tb_remove();
	window.location.reload();
}

async function enableSaveLocationIconButton() {
	const saveButtonElement = document.getElementById( 'save-location-icon' );
	if ( ! ( saveButtonElement instanceof HTMLElement ) ) {
		throw new Error( 'No save button found to enable!' );
	}

	saveButtonElement.disabled = false;

	const locationNoticeElement = document.getElementById( 'location-notice' );
	if ( locationNoticeElement instanceof HTMLElement ) {
		locationNoticeElement.innerHTML = '';
	}
}

async function disableSaveLocationIconButton() {
	const saveButtonElement = document.getElementById( 'save-location-icon' );
	if ( ! ( saveButtonElement instanceof HTMLElement ) ) {
		throw new Error( 'No save button found to enable!' );
	}

	saveButtonElement.disabled = true;

	const locationNoticeElement = document.getElementById( 'location-notice' );
	if ( locationNoticeElement instanceof HTMLElement ) {
		locationNoticeElement.innerHTML = '';
	}
}

async function geocodeAddress() {
	if ( ! map ) {
		throw new Error( 'No initialized map found!' );
	}

	const addressData = getAddressData();

	if ( ! addressData.country ) {
		throw new Error( 'No country selected!' );
	}

	const locationParams = new URLSearchParams();
	locationParams.set( 'size', '1' ); // maximum 1 result!
	let locationAddress = '';
	if ( addressData.street ) {
		locationAddress = addressData.street;
	}
	if ( addressData.house && locationAddress !== '' ) {
		locationAddress = `${ addressData.house } ${ locationAddress }`;
	}
	if ( locationAddress !== '' ) {
		locationParams.set( 'address', locationAddress );
	}
	if ( addressData.city ) {
		locationParams.set( 'locality', addressData.city );
	}
	if ( addressData.postal ) {
		locationParams.set( 'postalcode', addressData.postal );
	}
	if ( addressData.state ) {
		locationParams.set( 'region', addressData.state );
	}
	locationParams.set( 'country', addressData.country );

	try {
		const locationUrl = `https://geocode.locr.com/api/pelias/search/structured?${ locationParams.toString() }`;
		const locationResponse = await fetch( locationUrl );
		return await locationResponse.json();
	} catch ( exc ) {
		throw new Error( 'Error while geocoding address!' );
	}
}

/**
 * @return {{country?: string, state?: string, city?: string, postal?: string, street?: string, house?: string}} addressData
 */
function getAddressData() {
	const data = {};

	const streetElement = document.getElementById( 'location-address-street' );
	if ( streetElement && streetElement.value.trim() !== '' ) {
		data.street = streetElement.value.trim();
	}
	const houseElement = document.getElementById( 'location-address-house' );
	if ( houseElement && houseElement.value.trim() !== '' ) {
		data.house = houseElement.value.trim();
	}
	const cityElement = document.getElementById( 'location-address-city' );
	if ( cityElement && cityElement.value.trim() !== '' ) {
		data.city = cityElement.value.trim();
	}
	const postalElement = document.getElementById( 'location-address-postal' );
	if ( postalElement && postalElement.value.trim() !== '' ) {
		data.postal = postalElement.value.trim();
	}
	const stateElement = document.getElementById( 'location-address-state' );
	if ( stateElement && stateElement.value.trim() !== '' ) {
		data.state = stateElement.value.trim();
	}
	const countryElement = document.getElementById(
		'location-address-country'
	);
	if ( countryElement && countryElement.value.trim() !== '' ) {
		data.country = countryElement.value.trim();
	}

	return data;
}

/**
 * @param {{country?: string, state?: string, city?: string, postal?: string, street?: string, house?: string}} addressData
 */
function setAddressData( addressData ) {
	const streetElement = document.getElementById( 'location-address-street' );
	if ( streetElement && addressData.street ) {
		streetElement.value = addressData.street;
	}
	const houseElement = document.getElementById( 'location-address-house' );
	if ( houseElement && addressData.house ) {
		houseElement.value = addressData.house;
	}
	const cityElement = document.getElementById( 'location-address-city' );
	if ( cityElement && addressData.city ) {
		cityElement.value = addressData.city;
	}
	const postalElement = document.getElementById( 'location-address-postal' );
	if ( postalElement && addressData.postal ) {
		postalElement.value = addressData.postal;
	}
	const stateElement = document.getElementById( 'location-address-state' );
	if ( stateElement && addressData.state ) {
		stateElement.value = addressData.state;
	}
	const countryElement = document.getElementById(
		'location-address-country'
	);
	if ( countryElement && addressData.country ) {
		countryElement.value = addressData.country;
		if ( countryElement.value === '' ) {
			countryElement.value = '';
		}
	}
}

async function getSelectedLocationIcon() {
	const selectedLocationIconElement = document.querySelector(
		'#location-icons .selected img[data-meta]'
	);
	if ( ! ( selectedLocationIconElement instanceof HTMLElement ) ) {
		throw new Error( 'No location icon selected!' );
	}

	try {
		return new Icon(
			selectedLocationIconElement.getAttribute( 'data-meta' )
		);
	} catch ( exc ) {
		let errorMessage =
			'Error while parsing selected location icon meta-data!';
		if ( exc instanceof Error ) {
			errorMessage += ' Message: ' + exc.message;
		}
		throw new Error( errorMessage );
	}
}

function handleExceptionAsLocationNotice( exc ) {
	const saveLocationIcon = document.getElementById( 'save-location-icon' );
	if ( saveLocationIcon instanceof HTMLElement ) {
		saveLocationIcon.disabled = true;
	}

	if ( ! ( exc instanceof Error ) ) {
		return;
	}

	const locationNoticeElement = document.getElementById( 'location-notice' );
	if ( ! ( locationNoticeElement instanceof HTMLElement ) ) {
		return;
	}

	locationNoticeElement.innerHTML = '';

	const noticeElement = document.createElement( 'div' );
	noticeElement.classList.add( 'notice' );
	noticeElement.classList.add( 'notice-error' );
	noticeElement.classList.add( 'is-dismissible' );
	noticeElement.textContent = exc.message;
	locationNoticeElement.appendChild( noticeElement );
}

/**
 * @param {HTMLElement} element
 */
async function loadMap( element ) {
	const settings = await Settings.init( element );

	if ( ! map ) {
		map = window.L.map( element, { minZoom: 0, maxZoom: 19 } );
		map.attributionControl.setPrefix( '' );
		window.L.gridLayer
			.vms2( {
				attribution:
					'&copy; <a href="https://maps.locr.com">LOCR</a>/<a href="https://osm.org/copyright">OpenStreetMap</a>/<a href="https://leafletjs.com/">Leaflet</a>',
				accessKey: settings.accessKey,
				style: settings.style,
			} )
			.addTo( map );
	}

	map.setView(
		[ settings.position.latitude, settings.position.longitude ],
		settings.zoom
	);
}

function setSelectedLocationIconByIconId( iconId ) {
	const locationIconsElement = document.getElementById( 'location-icons' );
	if ( ! ( locationIconsElement instanceof HTMLElement ) ) {
		throw new Error( 'No location icons found!' );
	}

	const imgElements =
		locationIconsElement.querySelectorAll( 'img[data-meta]' );
	imgElements.forEach( ( imgElement ) => {
		imgElement.parentElement.classList.remove( 'selected' );
	} );
	for ( const imgElement of imgElements ) {
		try {
			const meta = JSON.parse( imgElement.getAttribute( 'data-meta' ) );
			if ( meta.id === iconId ) {
				imgElement.parentElement.classList.add( 'selected' );
			}
		} catch ( exc ) {
			continue;
		}
	}
}

async function setSelectedLocationIconToTheMap( locationResult ) {
	const selectedLocationIcon = await getSelectedLocationIcon();

	if ( marker ) {
		marker.remove();
	}
	marker = await buildMarkerFromLocationResultAndLocationIcon(
		locationResult,
		selectedLocationIcon
	);
	marker.addTo( map );
}

/**
 * @param {{bbox?: Array<number>, features?: Array<{geometry: {type: string, coordinates: Array<number>}}>}} locationResult
 */
async function setMapViewFromLocationResult( locationResult ) {
	let mapBoundsAlreadySet = false;
	if ( locationResult.bbox instanceof Array ) {
		map.fitBounds(
			L.latLngBounds(
				L.latLng( locationResult.bbox[ 1 ], locationResult.bbox[ 0 ] ),
				L.latLng( locationResult.bbox[ 3 ], locationResult.bbox[ 2 ] )
			),
			{ maxZoom: 19 }
		);
		mapBoundsAlreadySet = true;
	}

	if (
		locationResult.features instanceof Array &&
		locationResult.features.length > 0
	) {
		const location = locationResult.features[ 0 ];
		if ( location.geometry.type === 'Point' ) {
			if ( ! mapBoundsAlreadySet ) {
				map.setView(
					[
						location.geometry.coordinates[ 1 ],
						location.geometry.coordinates[ 0 ],
					],
					16
				);
			}
		}
	}

	return locationResult;
}

/**
 * @param {Icon} locationIcon
 */
async function tryUpdateLocationIconOnTheMap( locationIcon ) {
	if ( ! map || ! marker ) {
		return;
	}

	const originalMarkerPosition = marker.getLatLng();
	marker.remove();

	const locationResult = {
		features: [
			{
				geometry: {
					coordinates: [
						originalMarkerPosition.lng,
						originalMarkerPosition.lat,
					],
					type: 'Point',
				},
			},
		],
	};
	marker = await buildMarkerFromLocationResultAndLocationIcon(
		locationResult,
		locationIcon
	);
	marker.addTo( map );
}
